import { __assign, __spreadArrays } from "tslib";
import Ajv from 'ajv';
import ajvKeywords from 'ajv-keywords';
import cloneDeep from 'lodash/cloneDeep';
import defaults from 'lodash/defaults';
import every from 'lodash/every';
import findIndex from 'lodash/findIndex';
import findKey from 'lodash/findKey';
import map from 'lodash/map';
import pickBy from 'lodash/pickBy';
import reduce from 'lodash/reduce';
import startsWith from 'lodash/startsWith';
import trimStart from 'lodash/trimStart';
import { randomString, regexEscape } from '../../utils';
import { getDataModel } from '../DataTypes';
var ajv = new Ajv();
ajvKeywords(ajv, ['regexp', 'formatMaximum', 'formatMinimum']);
export var FULL_TEXT_SLUG = 'full_text_search';
var DEFAULT_DELIMITER = '___';
export var filter = function (filters, collection) {
    ajv.removeSchema(/^.*$/);
    var validators = Array.isArray(filters)
        ? filters.map(function (s) { return ajv.compile(s); })
        : [ajv.compile(filters)];
    if (Array.isArray(collection)) {
        return collection.filter(function (m) { return every(validators, function (v) { return v(m); }); });
    }
    return pickBy(collection, function (m) { return every(validators, function (v) { return v(m); }); });
};
export var createFullTextSearchFilter = function (schema, term) {
    var items = Object.entries(schema.properties).reduce(function (itemsAccumulator, entry) {
        var key = entry[0], item = entry[1];
        var typeArray = Array.isArray(item.type) ? item.type : [item.type];
        if (typeArray.includes('string') || typeArray.includes('number')) {
            itemsAccumulator.push({
                key: key,
                type: typeArray.includes('number') ? 'number' : 'string',
                numberProperties: typeArray.includes('number') && !isNaN(Number(term))
                    ? {
                        maximum: Number(term),
                        minimum: Number(term),
                    }
                    : null,
            });
        }
        return itemsAccumulator;
    }, []);
    var filteredItems = isNaN(Number(term))
        ? items.filter(function (i) { return i.type !== 'number'; })
        : items;
    var filter = {
        $id: randomString(),
        title: FULL_TEXT_SLUG,
        anyOf: [
            {
                description: "Any field contains " + term,
                anyOf: filteredItems.map(function (item) {
                    var _a;
                    return ({
                        properties: (_a = {},
                            _a[item.key] = __assign({ type: item.type, regexp: {
                                    pattern: regexEscape(term),
                                    flags: 'i',
                                } }, item.numberProperties),
                            _a),
                        required: [item.key],
                    });
                }),
            },
        ],
    };
    return filter;
};
export var upsertFullTextSearch = function (schema, filters, term) {
    if (filters === void 0) { filters = []; }
    var searchFilter = createFullTextSearchFilter(schema, term);
    var existingSearchIndex = findIndex(filters, { title: FULL_TEXT_SLUG });
    if (existingSearchIndex > -1) {
        return filters.map(function (filter, i) {
            return existingSearchIndex === i ? searchFilter : filter;
        });
    }
    return __spreadArrays(filters, [searchFilter]);
};
export var removeFullTextSearch = function (filters) {
    return filters.filter(function (f) { return f.title !== FULL_TEXT_SLUG; });
};
export var createFilter = function (schema, signatures) {
    var anyOf = signatures.map(function (_a) {
        var field = _a.field, operator = _a.operator, value = _a.value;
        var subSchema = schema.properties[field];
        var model = getDataModel(subSchema);
        if (!model || !model.operators.hasOwnProperty(operator)) {
            return {};
        }
        return model.createFilter(field, operator, value, subSchema);
    });
    return {
        $id: randomString(),
        anyOf: anyOf,
    };
};
export var decodeFilter = function (schema, filter) {
    var signatures = filter.anyOf.map(function (f) {
        if (!f) {
            return null;
        }
        if (!f.properties &&
            (!f.anyOf || !f.anyOf.length || !f.anyOf[0].properties)) {
            return null;
        }
        var schemaField = f.anyOf
            ? Object.keys(f.anyOf[0].properties).shift()
            : Object.keys(f.properties).shift();
        var subSchema = schema.properties[schemaField];
        var model = getDataModel(subSchema);
        if (!model) {
            return {};
        }
        return model.decodeFilter(f);
    });
    return signatures.filter(function (s) { return s !== null; });
};
export var getOperators = function (schema, field) {
    var subSchema = schema.properties[field];
    var model = getDataModel(subSchema);
    if (!model) {
        return [];
    }
    return map(model.operators, function (value, key) { return ({
        slug: key,
        label: value.getLabel(subSchema),
    }); });
};
var flattenAccumulator = function (schema, delimiter, parentKey, accumulator) {
    if (parentKey === void 0) { parentKey = ''; }
    if (!accumulator) {
        accumulator = cloneDeep(schema);
        if (schema.properties) {
            accumulator.properties = {};
        }
    }
    if (schema.anyOf && Array.isArray(schema.anyOf)) {
        accumulator.anyOf = schema.anyOf.map(function (subSchema) {
            return flattenAccumulator(subSchema, delimiter, parentKey);
        });
    }
    return reduce(schema.properties, function (result, value, key) {
        if (typeof value === 'boolean') {
            return accumulator;
        }
        var newKey = parentKey + delimiter + key;
        if (value.type !== 'object' ||
            (findKey(value.properties, { description: 'key' }) &&
                findKey(value.properties, { description: 'value' }))) {
            if (parentKey) {
                result.properties[newKey] = defaults(value, { title: key });
                return result;
            }
            result.properties[key] = value;
            return result;
        }
        return flattenAccumulator(value, delimiter, newKey, result);
    }, accumulator);
};
export var flattenSchema = function (schema, delimiter) {
    if (delimiter === void 0) { delimiter = DEFAULT_DELIMITER; }
    return flattenAccumulator(schema, delimiter);
};
export var unflattenSchema = function (schema, delimiter) {
    if (delimiter === void 0) { delimiter = DEFAULT_DELIMITER; }
    var base = cloneDeep(schema);
    if (schema.properties) {
        base.properties = {};
    }
    var unflattenedSchema = reduce(schema.properties, function (result, value, field) {
        if (!startsWith(field, delimiter)) {
            result.properties[field] = value;
            return result;
        }
        var keys = trimStart(field, delimiter).split(delimiter);
        var lastKey = keys.pop();
        var head = result;
        for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
            var key = keys_1[_i];
            if (!head.properties[key]) {
                head.properties[key] = {
                    type: 'object',
                    properties: {},
                };
            }
            head = head.properties[key];
        }
        head.properties[lastKey] = value;
        return result;
    }, base);
    if (unflattenedSchema.anyOf) {
        unflattenedSchema.anyOf = unflattenedSchema.anyOf.map(function (subSchema) {
            return unflattenSchema(subSchema, delimiter);
        });
    }
    if (unflattenedSchema.required) {
        var requiredFields = unflattenedSchema.required;
        unflattenedSchema.required = [];
        requiredFields.forEach(function (requiredField) {
            var fields = trimStart(requiredField, delimiter).split(delimiter);
            var head = unflattenedSchema;
            for (var _i = 0, fields_1 = fields; _i < fields_1.length; _i++) {
                var field = fields_1[_i];
                if (!head.required) {
                    head.required = [];
                }
                head.required.push(field);
                head = head.properties[field];
            }
        });
    }
    return unflattenedSchema;
};
export var getCleanEditModel = function (schema, field) {
    if (!field) {
        field = Object.keys(schema.properties).shift();
    }
    var fieldOperators = getOperators(schema, field);
    if (!fieldOperators.length) {
        return {
            field: field,
            operator: '',
            value: '',
        };
    }
    var operator = fieldOperators.shift().slug;
    var value = '';
    var subschema = schema.properties[field];
    if (typeof subschema !== 'boolean') {
        if (subschema.enum) {
            value = subschema.enum[0] || '';
        }
        if (subschema.oneOf) {
            value = subschema.oneOf[0].const || '';
        }
        if (subschema.type === 'boolean') {
            value = true;
        }
    }
    return {
        field: field,
        operator: operator,
        value: value,
    };
};
